/* This file is part of tagr.
   Copyright (C) 2000, 2005, 2009 Max Bouglacoff, Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free
#include <obstack.h>
#include <tagr.h>

#define ARG_UNUSED __attribute__ ((__unused__))

static struct obstack sd_stack;
static SD *sd_base;
static size_t sd_count;

static struct obstack pp_cmd_stack;
static int pp_cmd_stack_init;

SD *
find_router (const char *name)
{
  SD *sd;
  
  for (sd = sd_base; sd < sd_base + sd_count; sd++)
    if (strcmp (sd->name, name) == 0)
      return sd;
  return NULL; 
}

SD *
find_router_id (const char *id)
{
  SD *sd;

  for (sd = sd_base; sd < sd_base + sd_count; sd++)
    if (strcmp (sd->id, id) == 0)
      return sd;
  return NULL; 
}


static int
cb_router (enum grecs_callback_command cmd,
	   grecs_locus_t *locus,
	   void *varptr,
	   grecs_value_t *value,
	   void *cb_data)
{
  int rc;
  SD *sd;
  void **pdata = cb_data;
  
  switch (cmd)
    {
    case grecs_callback_section_begin:
      if (!value || value->type != GCONF_TYPE_STRING)
	{
	  grecs_error (locus, 0, _("tag must be a string"));
	  return 0;
	}
      sd = xzalloc (sizeof (*sd));
      sd->id = strdup (value->v.string);
      *pdata = sd;
      break;

    case grecs_callback_section_end:
      sd = *pdata;
      if (sd->max == 0)
	{
	  grecs_error (locus, 0, _("maximum speed not set"));
	  free (sd);
	  return 0;
	}
      if (!sd->name)
	sd->name = strdup (sd->id);
      if (!sd->dir)
	sd->dir = strdup (sd->id);
      if (sd_count == 0)
	obstack_init (&sd_stack);

      obstack_grow (&sd_stack, sd, sizeof *sd);
      sd_count++;
      free (sd);
      *pdata = NULL;
      break;

    case grecs_callback_set_value:
      grecs_error (locus, 0, _("invalid use of block statement"));
  }
  return 0;
}


static struct grecs_keyword router_kw[] = {
  { "host", NULL, N_("Host name or IP address"),
    grecs_type_string, NULL, offsetof(struct HostData, name) },
  { "directory", N_("name"), N_("Subdirectory name"),
    grecs_type_string, NULL, offsetof(struct HostData, dir) },
  { "max-speed", NULL, N_("Maximum speed"),
    grecs_type_size, NULL, offsetof(struct HostData, max) },
  { NULL }
};


static int
cb_number_suffixes (enum grecs_callback_command cmd,
		    grecs_locus_t *locus,
		    void *varptr,
		    grecs_value_t *value,
		    void *cb_data)
{
  switch (value->type)
    {
    case GCONF_TYPE_STRING:
        number_suffix_count = 2;
	number_suffix = xcalloc (2, sizeof (number_suffix[0]));
	number_suffix[0] = "";
	number_suffix[1] = value->v.string;
	break;

    case GCONF_TYPE_ARRAY:
      {
	int i;
	
	number_suffix_count = value->v.arg.c + 1;
	number_suffix = xcalloc (number_suffix_count,
				 sizeof (number_suffix[0]));
	number_suffix[0] = "";
      
	for (i = 0; i < value->v.arg.c; i++)
	  {
	    grecs_value_t *pv = &value->v.arg.v[i];
	    if (pv->type != GCONF_TYPE_STRING)
	      {
		grecs_error (locus, 0,
			     _("expected scalar value but found list"));
		return 1;
	      }
	    number_suffix[i] = pv->v.string;
	  }
      }
      break;
	
    case GCONF_TYPE_LIST:
      {
	const void *p;
	int i = 1;
	gl_list_iterator_t itr = gl_list_iterator (value->v.list);
	
	while (gl_list_iterator_next (&itr, &p, NULL))
	  {
	    const grecs_value_t *vp = p;
	    if (vp->type != GCONF_TYPE_STRING)
	      {
		grecs_error (locus, 0,
			     _("expected scalar value but found list"));
		return 1;
	      }
	    number_suffix[i++] = vp->v.string;
	  }
	gl_list_iterator_free (&itr);
      }
    }
  return 0;
}



static struct grecs_keyword tagr_kw[] = {
  { "basedir", NULL, N_("Set base directory name"),
    grecs_type_string, &basedir },
  { "user", NULL, N_("Run with this user privileges"),
    grecs_type_string, &user },
  { "template", NULL, N_("HTML page template"),
    grecs_type_string, &html_template },
  { "pidfile", NULL, N_("Set pid file name"),
    grecs_type_string, &pidfile },
  { "listen", N_("socket"), N_("Listen on this address"),
    grecs_type_sockaddr, &listen_sockaddr, },

  { "router", N_("id: string"), N_("Configure a router"),
    grecs_type_section, NULL, 0,
    cb_router, NULL, router_kw },

  { "number-suffixes", N_("suffixes"), NULL /* FIXME */,
    grecs_type_string, NULL, 0, cb_number_suffixes },
  
  { "transparent", NULL, N_("Transparent graphs"),
    grecs_type_bool, &transparent_option },
  { "percent", NULL, NULL /* FIXME: descr */,
    grecs_type_bool, &percent_option },
  { "zero-unknown", NULL, NULL /* FIXME: descr */,
    grecs_type_bool, &zero_unknown_option },
  { "fill-incoming", NULL, NULL /* FIXME: descr */,
    grecs_type_bool, &fill_incoming_option },

  { "color-background", NULL, N_("Set background color"),
    grecs_type_int, &color_background },
  { "color-light", NULL, NULL /* FIXME: descr */,
    grecs_type_int, &color_light },
  { "color-dark", NULL, NULL /* FIXME: descr */,
    grecs_type_int, &color_dark },
  { "color-major", NULL, NULL /* FIXME: descr */,
    grecs_type_int, &color_major },
  { "color-in", NULL, NULL /* FIXME: descr */,
    grecs_type_int, &color_in },
  { "color-out", NULL, NULL /* FIXME: descr */,
    grecs_type_int, &color_out },
  { "color-grid", NULL, NULL /* FIXME: descr */,
    grecs_type_int, &color_grid },
  { "color-in-max", NULL, NULL /* FIXME: descr */,
    grecs_type_int, &color_in_max },
  { "color-out-max", NULL, NULL /* FIXME: descr */,
    grecs_type_int, &color_out_max },
  { "color-percent", NULL, NULL /* FIXME: descr */,
    grecs_type_int, &color_percent },
  
  /* FIXME */
  { NULL }
};

void
config_help ()
{
    static char docstring[] =
      N_("Configuration file structure for tagr.\n");
    /* FIXME:
       "For more information, use `info tagr configuration'."); */
    grecs_format_docstring (stdout, docstring, 0);
    grecs_format_statement_array (stdout, tagr_kw, 1, 0);
}

int
readconfig ()
{
  int rc;
  
  grecs_set_keywords (tagr_kw);
  grecs_include_path_setup (DEFAULT_VERSION_INCLUDE_DIR,
			    DEFAULT_INCLUDE_DIR, NULL);
  grecs_preprocessor = DEFAULT_PREPROCESSOR;
  grecs_log_to_stderr = log_to_stderr;

  if (pp_cmd_stack_init && grecs_preprocessor)
    {
      char *defs = obstack_finish (&pp_cmd_stack);
      char *cmd = xmalloc (strlen (grecs_preprocessor) + strlen (defs) + 1);
      strcpy (cmd, grecs_preprocessor);
      strcat (cmd, defs);
      grecs_preprocessor = cmd;
      obstack_free (&pp_cmd_stack, NULL);
    }
  if (preprocess_only)
    exit (grecs_preproc_run (configfile, grecs_preprocessor) ? EX_CONFIG : 0);
  rc = grecs_parse (configfile);
  if (rc == 0)
    sd_base = obstack_finish (&sd_stack);
  return rc;
}

void
define_symbol (char *p)
{
  if (!pp_cmd_stack_init)
    {
      obstack_init (&pp_cmd_stack);
      pp_cmd_stack_init = 1;
    }
  obstack_grow (&pp_cmd_stack, " \"-D", 4);
  for (p = optarg; *p; p++)
    {
      if (*p == '\\' || *p == '"')
	obstack_1grow (&pp_cmd_stack, '\\');
      obstack_1grow (&pp_cmd_stack, *p);
    }
  obstack_1grow (&pp_cmd_stack, '"');  
}
